@@ -22,6 +22,8 @@ SOFTWARE.
 
 #include "multivariate-pdf.h"
 #include <R.h>
+#include <R_ext/BLAS.h>
+#include <R_ext/Lapack.h>
 #include <R_ext/Memory.h>
 #include <R_ext/RS.h> /* for R_Calloc/R_Free, F77_CALL */
 #include <Rinternals.h>
@@ -41,34 +43,41 @@ SOFTWARE.
 typedef struct
 {
     Matrix muR;      // G x (C-1)
-    Matrix **sigma;  // array G of (C-1 x C-1) matrices
-    double *feature; // size C (temp: X[,b])
-    double *muG;     // size C-1 (temp: muR[g,])
-    double *QC;      // size C (per-group numerators)
-    double *maha;    // size G*C (mahalanobis distances, row-major)
+    Matrix **sigma;  // array G de matrices (C-1 x C-1)
+    double *feature; // size C
+    double *muG;     // size C-1
+    double *QC;      // size C
+    double *maha;    // size G*C
+    // scratch:
+    double *diff;      // size C-1
+    double *y;         // size C-1
+    double *z;         // size C-1
+    double *ec;        // size C-1
+    double *diag_Sinv; // size G*(C-1)
 } Arena;
 
-// Create and initialize an Arena
 static Arena Arena_init(int G, int C)
 {
-    Arena A = {0};
+    Arena A = (Arena){0};
     A.muR = createMatrix(G, C - 1);
-
     A.sigma = (Matrix **)Calloc(G, Matrix *);
     for (int g = 0; g < G; ++g)
     {
         A.sigma[g] = (Matrix *)Calloc(1, Matrix);
         *(A.sigma[g]) = createMatrix(C - 1, C - 1);
     }
-
     A.feature = (double *)Calloc(C, double);
     A.muG = (double *)Calloc(C - 1, double);
     A.QC = (double *)Calloc(C, double);
     A.maha = (double *)Calloc((size_t)G * (size_t)C, double);
 
+    A.diff = (double *)Calloc(C - 1, double);
+    A.y = (double *)Calloc(C - 1, double);
+    A.z = (double *)Calloc(C - 1, double);
+    A.ec = (double *)Calloc(C - 1, double);
+    A.diag_Sinv = (double *)Calloc((size_t)G * (size_t)(C - 1), double); // opcional
     return A;
 }
-
 // Frees an Arena
 static void Arena_free(Arena *A, int G)
 {
@@ -93,6 +102,16 @@ static void Arena_free(Arena *A, int G)
         Free(A->QC);
     if (A->maha)
         Free(A->maha);
+    if (A->diff)
+        Free(A->diff);
+    if (A->y)
+        Free(A->y);
+    if (A->z)
+        Free(A->z);
+    if (A->ec)
+        Free(A->ec);
+    if (A->diag_Sinv)
+        Free(A->diag_Sinv);
     memset(A, 0, sizeof(*A));
 }
 
@@ -114,73 +133,137 @@ static inline void getRow_into(const Matrix *M, int row, double *out)
  * Given a ballot box index, probabilities and the reduced version (with C-1 candidates) of the probabilities matrix, it
  * calculates the `q` values in a flattened way
  *
- * @param[in] b. The index of the ballot box
- * @param[in] *probabilities. A pointer towards the probabilities matrix.
- * @param[in] *probabilitiesReduced. A pointer towards the reduced probabilities matrix.
- *
- * @return A (g x c) matrix with the values of `q` according the candidate and group index.
+ * @param[in] b The index of the ballot box
+ * @param[in] *probabilities A pointer towards the probabilities matrix
+ * @param[in] *probabilitiesReduced A pointer towards the reduced probabilities matrix
+ * @param[in,out] *ll Log-likelihood accumulator (only used if params.computeLL = true)
+ * @param[in] params Method input configuration
+ * @param[in,out] *A Arena scratch space
  *
+ * @return Writes into ctx->q a (B x G x C) tensor of q values
  */
 void computeQforABallot(EMContext *ctx, int b, const Matrix *probabilities, const Matrix *probabilitiesReduced,
                         double *ll, QMethodInput params, Arena *A)
 {
     const int G = (int)ctx->G;
     const int C = (int)ctx->C;
+    const int n = C - 1;
     Matrix *X = &ctx->X;
 
     // ---- Fill muR and sigma for this ballot ---- //
     getAverageConditional(ctx, b, probabilitiesReduced, &A->muR, A->sigma);
 
-    // Invert sigmas
-    for (int g = 0; g < G; ++g)
-        inverseSymmetricPositiveMatrix(A->sigma[g]);
-
-    // ---- Compute determinant (for loglikelihood normalization) ---- //
+    // ---- Normalization constant for log-likelihood ---- //
     double normalizeConstant = 1.0;
     if (params.computeLL)
     {
-        double det = 1.0;
-        for (int c = 0; c < C - 1; ++c)
-            det *= MATRIX_AT_PTR(A->sigma[0], c, c);
-        det = 1.0 / (det * det);
-        normalizeConstant = R_pow(R_pow_di(M_2_PI, C - 1) * det, 0.5);
+        if (C == 2)
+        {
+            // For C=2, we use the inverse later. Determinant can be computed if needed.
+            normalizeConstant = 1.0; // neutral factor
+        }
+        else
+        {
+            // For C > 2, sigma[g] stores the Cholesky factor L, so determinant = (prod diag(L))^2
+            double det = 1.0;
+            for (int c = 0; c < n; ++c)
+                det *= MATRIX_AT_PTR(A->sigma[0], c, c);
+            det = 1.0 / (det * det);
+            normalizeConstant = R_pow(R_pow_di(M_2_PI, n) * det, 0.5);
+        }
     }
 
-    // ---- Feature vector (candidate results) ----
+    // ---- Feature vector (candidate results) ---- //
     getColumn_into(X, b, A->feature);
 
-    // ---- Mahalanobis per group ----
-    for (int g = 0; g < G; ++g)
-    { // --- For each group
-        getRow_into(&A->muR, g, A->muG);
-        double *dst = &A->maha[(size_t)g * (size_t)C];
-        // --- Mahalanobis distance for each candidate given a group, accounting the mean
-        getMahanalobisDist(A->feature, A->muG, A->sigma[g], dst, C - 1, false);
+    // ---- Sigma preparation depending on C ---- //
+    if (C == 2)
+    {
+        // Inverse Sigma directly for Mahalanobis original implementation
+        for (int g = 0; g < G; ++g)
+        {
+            inverseSymmetricPositiveMatrix(A->sigma[g]); // now A->sigma[g] = Sigma^{-1}
+        }
+    }
+    else
+    {
+        // Cholesky factorization for Mahalanobis stable computation
+        for (int g = 0; g < G; ++g)
+        {
+            choleskyMat(A->sigma[g]); // store L (lower) in A->sigma[g]
+        }
     }
 
-    // ---- build q’s directly into ctx->q ----
+    // ---- Mahalanobis distance per group ---- //
     for (int g = 0; g < G; ++g)
-    { // --- For each group
-        double den = 0.0;
+    {
+        getRow_into(&A->muR, g, A->muG); // \mu_g (size n)
+        for (int i = 0; i < n; ++i)
+            A->diff[i] = A->feature[i] - A->muG[i];
+
+        double *dst = &A->maha[(size_t)g * (size_t)C];
+
+        if (C == 2)
+        {
+            // Original implementation with inverse Sigma
+            getMahanalobisDist(A->feature, A->muG, A->sigma[g], dst, n, /*reduced*/ false);
+        }
+        else
+        {
+            // Current implementation with Cholesky
+            double *Sdiag_g = A->diag_Sinv ? &A->diag_Sinv[g * (size_t)n] : NULL;
+            double baseline = getMahanalobisDist2(A->sigma[g], A->diff, A->y, A->z, A->ec, Sdiag_g, n, /*need_z*/ true,
+                                                  /*need_diag*/ true);
+
+            // Last candidate = baseline
+            dst[n] = baseline;
+            for (int c = 0; c < n; ++c)
+            {
+                double diagcc = Sdiag_g ? Sdiag_g[c] : 0.0;
+                dst[c] = baseline - 2.0 * A->z[c] + diagcc;
+            }
+        }
+    }
+
+    // ---- Build q’s directly into ctx->q (always in log-space) ---- //
+    for (int g = 0; g < G; ++g)
+    {
         double *ma = &A->maha[(size_t)g * (size_t)C];
 
+        double logw[C];
+        double logw_max = -INFINITY;
+
+        // --- Numerators in log-space ---
         for (int c = 0; c < C; ++c)
-        { // --- For each candidate given a group
-          // ---- The `q` value is calculated as exp(-0.5 * mahanalobis) * probabilities ----
-            double num = exp(-0.5 * ma[c]) * MATRIX_AT_PTR(probabilities, g, c);
-            // ---- Store the numerator temporarily in the arena ----
-            A->QC[c] = num;
-            den += num;
+        {
+            double prior = MATRIX_AT_PTR(probabilities, g, c);
+            double logP = (prior > 0.0) ? log(prior) : -INFINITY;
+            logw[c] = -0.5 * ma[c] + logP; // exp(-0.5*ma[c]) * prior
+            if (isfinite(logw[c]) && logw[c] > logw_max)
+                logw_max = logw[c];
         }
 
-        if (g == 0 && params.computeLL && den > 0.0)
-            // Normalize and accumulate log-likelihood
-            *ll += log(den) * normalizeConstant;
-
+        // --- Softmax with shift ---
+        double den = 0.0;
         for (int c = 0; c < C; ++c)
-        { // --- For each candidate given a group
-            double qgc = (den != 0.0) ? (A->QC[c] / den) : 0.0;
-            Q_3D(ctx->q, b, g, c, G, C) = (!isnan(qgc) && !isinf(qgc)) ? qgc : 0.0;
+        {
+            double val = isfinite(logw[c]) ? exp(logw[c] - logw_max) : 0.0;
+            A->QC[c] = val;
+            den += val;
+        }
+
+        // --- Log-likelihood contribution ---
+        if (g == 0 && params.computeLL && den > 0.0 && isfinite(logw_max))
+        {
+            double logden = logw_max + log(den);
+            *ll += logden * log(normalizeConstant);
+        }
+
+        // --- Normalize and store q ---
+        for (int c = 0; c < C; ++c)
+        {
+            double qgc = (den > 0.0) ? (A->QC[c] / den) : 0.0;
+            Q_3D(ctx->q, b, g, c, G, C) = qgc;
         }
     }
 }
@@ -208,7 +291,7 @@ void computeQMultivariatePDF(EMContext *ctx, QMethodInput params, double *ll)
 
     for (uint32_t b = 0; b < ctx->B; ++b)
     {
-        computeQforABallot(ctx, (int)b, probabilities, &probabilitiesReduced, ll, params, &A);
+        computeQforABallot(ctx, b, probabilities, &probabilitiesReduced, ll, params, &A);
     }
 
     // ---- Free the arena ---- //
