@@ -18,7 +18,7 @@ extern "C"
      *
      * @return A pointer to a flattened 3D array (b x g x c) representing the probabilities.
      */
-    double *computeQMultivariatePDF(Matrix const *probabilities, QMethodInput params, double *ll);
+    void computeQMultivariatePDF(EMContext *ctx, QMethodInput params, double *ll);
 
 #ifdef __cplusplus
 }
